#include <stdio.h> // printf
#include <assert.h> // assert
#include <cuda.h> // CUDA library
#include "helper_cuda.h" // CUDA error checking

// 2D to 1D bijection
#define IX(i,j) ((i)*(N)+(j))
// integer ceiling division
#define DIV_CEIL(a,b) (((a)+(b)-1)/(b))
// maximum
#define MAX(a,b) (((a)>(b))?(a):(b))

__global__ void setmm(const uint N, float * __restrict__ a, float * __restrict__ b, float * __restrict__ c) {
	const uint i = blockIdx.y*blockDim.y + threadIdx.y;
	const uint j = blockIdx.x*blockDim.x + threadIdx.x;
	if (i<N && j<N) {
		a[IX(i,j)] = sinf((float)i*j);
		b[IX(i,j)] = cosf((float)i+j);
		c[IX(i,j)] = 0.0f;
	}
}

__global__ void sgemm_multi(const uint N, const uint MULTI, float * __restrict__ a, float * __restrict__ b, float * __restrict__ c) {
	const uint i = blockIdx.y*blockDim.y + threadIdx.y;
	const uint j = MULTI*(blockIdx.x*blockDim.x + threadIdx.x);
	if (i<N && j<N)
		for (uint jj=j; jj<j+MULTI; ++jj)
			for (uint k=0; k<N; ++k)
				c[IX(i,jj)] += a[IX(i,k)] * b[IX(k,jj)];
}

int main(int argc, char **argv)
{
	float *d_a = NULL, *h_a = NULL; // pointers to a, b and c in Host and Device
	float *d_b = NULL, *h_b = NULL;
	float *d_c = NULL, *h_c = NULL;
	assert(5 == argc);
	const unsigned int N = atoi(argv[1]); // mtx side size
	const unsigned int BX = atoi(argv[2]); // block columns
	const unsigned int BY = atoi(argv[3]); // block rows
	const unsigned int MULTI = atoi(argv[4]); // # cij cells per thread
	const unsigned int SIZE = N*N; // linear size

	checkCudaErrors(cudaMalloc(&d_a, SIZE * sizeof(float)));
	checkCudaErrors(cudaMalloc(&d_b, SIZE * sizeof(float)));
	checkCudaErrors(cudaMalloc(&d_c, SIZE * sizeof(float)));
	checkCudaErrors(cudaMallocHost(&h_a, SIZE * sizeof(float)));
	checkCudaErrors(cudaMallocHost(&h_b, SIZE * sizeof(float)));
	checkCudaErrors(cudaMallocHost(&h_c, SIZE * sizeof(float)));
	assert(d_a && d_b && d_c && h_a && h_b && h_c);

	dim3 grid_size_setmm(DIV_CEIL(N,BX), DIV_CEIL(N,BY));
	dim3 block_size_setmm(BX, BY);
	setmm<<<grid_size_setmm, block_size_setmm>>>(N, d_a, d_b, d_c);
	getLastCudaError("setmm kernel failed");
	dim3 grid_size_sgemm(DIV_CEIL(N,BX*MULTI), DIV_CEIL(N,BY));
	dim3 block_size_sgemm(BX,BY);
	sgemm_multi<<<grid_size_sgemm, block_size_sgemm>>>(N, MULTI, d_a, d_b, d_c);
	getLastCudaError("sgemm_multi kernel failed");
	cudaDeviceSynchronize();

	checkCudaErrors(cudaMemcpy(h_a, d_a, SIZE*sizeof(float), cudaMemcpyDefault));
	checkCudaErrors(cudaMemcpy(h_b, d_b, SIZE*sizeof(float), cudaMemcpyDefault));
	checkCudaErrors(cudaMemcpy(h_c, d_c, SIZE*sizeof(float), cudaMemcpyDefault));
#ifdef DEBUG
	double max_diff = 0.0;
	for (size_t i=0; i<N; ++i) {
		for (size_t j=0; j<N; ++j) {
			float cij = 0.0f;
			for (size_t k=0; k<N; ++k)
				cij += h_a[IX(i,k)] * h_b[IX(k,j)];
			max_diff = MAX(max_diff, abs(cij-h_c[IX(i,j)]));
		}
	}
	printf("max_diff: %f\n", max_diff);
#endif

	checkCudaErrors(cudaFreeHost(h_c));
	checkCudaErrors(cudaFreeHost(h_b));
	checkCudaErrors(cudaFreeHost(h_a));
	checkCudaErrors(cudaFree(d_c));
	checkCudaErrors(cudaFree(d_b));
	checkCudaErrors(cudaFree(d_a));

	return 0;
}
