#include <stdio.h> // printf
#include <stdlib.h> // calloc
#include <assert.h> // assert
#include <cuda.h> // CUDA library
#include "helper_cuda.h"

// 2D to 1D bijection
#define IX(i,j) ((i)*(N)+(j))
// maximum
#define MAX(a,b) (((a)>(b))?(a):(b))

#ifndef N
#define N 1024
#endif

#ifndef B
#define B 32
#endif

#ifndef U
#define U 16
#endif

#define SIZE (N*N)

__global__ void setmm(float * __restrict__ a, float * __restrict__ b, float * __restrict__ c) {
	const uint i = blockIdx.y*B + threadIdx.y;
	const uint j = blockIdx.x*B + threadIdx.x;
	a[IX(i,j)] = sinf((float)i*j);
	b[IX(i,j)] = cosf((float)i+j);
	c[IX(i,j)] = 0.0f;
}

__global__ void sgemm_shared(float * __restrict__ a, float * __restrict__ b, float * __restrict__ c) {
	__shared__ float sa[B][B];
	__shared__ float sb[B][B];

	const uint ti = threadIdx.y;
	const uint tj = threadIdx.x;
	const uint i = blockIdx.y*B + ti;
	const uint j = blockIdx.x*B + tj;
	float cij[U] = {0.0f};
	#pragma unroll 32
	for (uint m=0; m<N/B; ++m) {
		#pragma unroll 8
		for (uint u=0; u<U; ++u) {
			sa[ti+u*B/U][tj] = a[IX(i+u*B/U,m*B+tj)];
			sb[ti+u*B/U][tj] = b[IX(m*B+ti+u*B/U,j)];
		}
		__syncthreads();

		#pragma unroll 32
		for (uint k=0; k<B; ++k) {
			#pragma unroll 8
			for (uint u=0; u<U; ++u) {
				cij[u] += sa[ti+u*B/U][k] * sb[k][tj];
			}
		}
		__syncthreads();
	}
	#pragma unroll 8
	for (uint u=0; u<U; ++u)
		c[IX(i+u*B/U,j)] = cij[u];
}

int main(int argc, char **argv)
{
	float *d_a = NULL, *h_a = NULL; // pointers to a, b and c in Host and Device
	float *d_b = NULL, *h_b = NULL;
	float *d_c = NULL, *h_c = NULL;
	assert(1 == argc); // no args
	assert(0 == N%B); // evenly split
	assert(0 == B%U); // unroll factor

	checkCudaErrors(cudaMallocHost(&h_a, SIZE * sizeof(float)));
	checkCudaErrors(cudaMallocHost(&h_b, SIZE * sizeof(float)));
	checkCudaErrors(cudaMallocHost(&h_c, SIZE * sizeof(float)));
	checkCudaErrors(cudaMalloc(&d_a, SIZE * sizeof(float)));
	checkCudaErrors(cudaMalloc(&d_b, SIZE * sizeof(float)));
	checkCudaErrors(cudaMalloc(&d_c, SIZE * sizeof(float)));
	assert(d_a && d_b && d_c && h_a && h_b && h_c);

	dim3 grid_size(N/B, N/B);
	dim3 block_size(B,B);
	setmm<<<grid_size, block_size>>>(d_a, d_b, d_c);
	getLastCudaError("setmm kernel failed");
	block_size.y /= U; // ajusto porque hago U celdas por hilo
	sgemm_shared<<<grid_size, block_size>>>(d_a, d_b, d_c);
	getLastCudaError("sgemm kernel failed");
	cudaDeviceSynchronize();

	checkCudaErrors(cudaMemcpy(h_a, d_a, SIZE*sizeof(float), cudaMemcpyDefault));
	checkCudaErrors(cudaMemcpy(h_b, d_b, SIZE*sizeof(float), cudaMemcpyDefault));
	checkCudaErrors(cudaMemcpy(h_c, d_c, SIZE*sizeof(float), cudaMemcpyDefault));
#ifdef DEBUG
	double max_diff = 0.0;
	for (size_t i=0; i<N; ++i) {
		for (size_t j=0; j<N; ++j) {
			float cij = 0.0f;
			for (size_t k=0; k<N; ++k)
				cij += h_a[IX(i,k)] * h_b[IX(k,j)];
			max_diff = MAX(max_diff, abs(cij-h_c[IX(i,j)]));
		}
	}
	printf("max_diff: %f\n", max_diff);
#endif

	checkCudaErrors(cudaFreeHost(h_c));
	checkCudaErrors(cudaFreeHost(h_b));
	checkCudaErrors(cudaFreeHost(h_a));
	checkCudaErrors(cudaFree(d_c));
	checkCudaErrors(cudaFree(d_b));
	checkCudaErrors(cudaFree(d_a));

	return 0;
}
